# distutils: language = c++
# cython: language_level = 3
cimport numpy as np

np.import_array()


from runtime.runtime cimport BatchMailBox, NullResult, Scheduler
from libc.stdio cimport printf
from libc.stdlib cimport malloc, free

from stdlib.string cimport string

## Types declaration
ctypedef int I_t
ctypedef double D_t

cdef lock Scheduler scheduler

cdef extern from *:
    """
    #include <algorithm>
    template<class D, class I>
    class IndexComparator {
    private:
        const D *data;
        I split_dim, n_features;
    public:
        IndexComparator(const D *data, const I &split_dim, const I &n_features):
            data(data), split_dim(split_dim), n_features(n_features) {}
        bool operator()(const I &a, const I &b) const {
            D a_value = data[a * n_features + split_dim];
            D b_value = data[b * n_features + split_dim];
            return a_value == b_value ? a < b : a_value < b_value;
        }
    };
    template<class D, class I>
    void partition_node_indices(
        const D *data,
        I *node_indices,
        const I &split_dim,
        const I &split_index,
        const I &n_features,
        const I &n_points) {
        IndexComparator<D, I> index_comparator(data, split_dim, n_features);
        std::nth_element(
            node_indices,
            node_indices + split_index,
            node_indices + n_points,
            index_comparator);
    }
    """
    void partition_node_indices[D, I](
                D *data,
                I *node_indices,
                I split_dim,
                I split_index,
                I n_features,
                I n_points) nogil except +

cdef cypclass Node activable:
    """A KDTree Node"""

    D_t * point
    I_t n_dims
    active Node left
    active Node right

    __init__(self):
        self._active_result_class = NullResult
        self._active_queue_class = consume BatchMailBox(scheduler)
        self.left = NULL
        self.right = NULL

    void build_node(
            self,
            D_t * points,
            I_t * indices,
            I_t depth,
            I_t n_dims,
            I_t dim,
            I_t start,
            I_t end,
    ):
        cdef I_t i
        cdef I_t next_dim = (dim + 1) % n_dims
        cdef I_t nn = end - start
        cdef I_t n_mid = nn // 2
        cdef I_t split_index = (start + end) // 2

        self.n_dims = n_dims

        if (depth < 0) or (nn <= 1):
            return

        printf("Depth %d on dim %d: [%d, %d) med: %d\n\n", depth,
        dim, start, end, split_index)

        partition_node_indices(points + start,
                indices + start, dim, n_mid, n_dims, nn)
        self.point = points + split_index


        self.left = activate(consume Node())
        self.right = activate(consume Node())

        self.left.build_node(NULL,
                             points, indices,
                             depth = depth - 1,
                             n_dims=n_dims,
                             dim=next_dim,
                             start=start, end=split_index)

        self.right.build_node(NULL,
                              points, indices,
                              depth=depth - 1,
                              n_dims=n_dims,
                              dim=next_dim,
                              start=split_index, end=end)


cdef cypclass KDTree:
    """A KDTree based on asynchronous and parallel computations.
    
    It is similar to :class:`sklearn.neighbours.KDTree` but it is 
    using Cython+ features for asynchronous computations.
    
    Asynchronuous and parallel computations can be used when 
    constructing the nodes (see above) and when querying (WIP).
    
    This relies on a Cython+ runtime using actors.
    """

    I_t n # number of points
    I_t d # number of dimensions / features
    I_t depth # max_depth of the tree (to be unified with leaf_size)

    np.ndarray data_arr
    np.ndarray idx_array_arr

    active Node root

    D_t *points
    I_t *indices

    __init__(self,
             np.ndarray data,
             I_t depth,
             ):
        cdef I_t i
        cdef I_t n = data.shape[0]
        cdef I_t d = data.shape[1]

        self.n = n
        self.d = d
        self.depth = depth

        # TODO: define it on the front-end
        # Use Golden Spiral for the layout
        cdef D_t golden_ratio = (1 + 5**0.5)/2
        self.points = <D_t *> malloc(n * d * sizeof(D_t))
        self.indices = <I_t *> malloc(n * sizeof(I_t))

        for i in range(n):
            self.indices[i] = i
            self.points[i * d] = (i / golden_ratio) % 1
            self.points[i * d + 1] = i / n

        # TODO: end

        self._recursive_build()

    void _recursive_build(self):
        # TODO: introducing a context for the runtime
        # would be nice here:
        #   ```
        #   with scheduler:
        #        self.root = ...
        #   ```
        global scheduler
        scheduler = Scheduler()

        self.root = activate(consume Node())
        if self.root is NULL:
            printf("Error consuming node\n")

        # When object are activated (set as Actors), methods
        # are reified. When using those reified methods
        # a new argument is prepredend for the Promise,
        # which we aren't using using here, hence the extra NULL.
        self.root.build_node(NULL,
                             self.points,
                             self.indices,
                             depth, d, dim=0, start=0, end=n)

        scheduler.finish()
        del scheduler

        for i in range(n):
            printf("indices[%d] = %d\n", i, self.indices[i])

    void __dealloc__(self):
        printf("Deallocating KDTree datastructures\n")
        free(self.points)
        free(self.indices)
        printf("Done deallocating KDTree datastructures\n")

cdef public int main() nogil:
    # Entry point for the compiled binary file
    tree = KDTree()
    printf("Done\n")
    # XXX a segfault is thrown when exiting
    # this function, but not others
    return 0

def python_main():
    tree = KDTree()

